/*
    ext2_resize.c -- ext2 resizer
    Copyright (C) 1998,99 Lennert Buytenhek <buytenh@dsv.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

static const char _ext2_resize_c[] = "$Id: ext2_resize.c,v 1.39 1999/07/26 09:46:54 buytenh Exp $";

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include "ext2.h"

static void ext2_add_group(struct ext2_fs *fs, blk_t groupsize)
{
	blk_t admin;
	int   i;
	int   group;
	blk_t groupstart;
	blk_t newgdblocks;
	int   sparse;

	fprintf(stderr, "ext2_add_group\n");

	if (fs->sb.s_blocks_count != fs->sb.s_first_data_block + fs->numgroups * fs->sb.s_blocks_per_group)
	{
		fprintf(stderr, "ext2_add_group: last (existing) group isn't complete!\n");
		return;
	}

	group = fs->numgroups;
	sparse = ext2_is_group_sparse(fs, group);
	groupstart = fs->sb.s_first_data_block + group * fs->sb.s_blocks_per_group;

	admin = fs->adminblocks;
	if (!sparse)
		admin -= fs->gdblocks + 1;

	if (groupsize < fs->adminblocks || groupsize > fs->sb.s_blocks_per_group)
	{
		fprintf(stderr, "ext2_add_group: groups of %i blocks are impossible!\n", groupsize);
		return;
	}

	newgdblocks = howmany((fs->numgroups + 1) * sizeof(struct ext2_group_desc), fs->blocksize);
	if (newgdblocks != fs->gdblocks)
	{
		for (i=0;i<fs->numgroups;i++)
		{
			int fromblock;

			fromblock = fs->gd[i].bg_inode_table + fs->inodeblocks;
			if (ext2_get_block_state(fs, fromblock))
			{
				fprintf(stderr, "error: block relocator should have relocated %i\n", fromblock);
				return;
			}

			fs->gd[i].bg_free_blocks_count--;
			fs->sb.s_free_blocks_count--;

			ext2_set_block_state(fs, fromblock, 1, 0);

			while (fromblock-- > fs->gd[i].bg_inode_table)
				ext2_copy_block(fs, fromblock, fromblock+1);
			fs->gd[i].bg_inode_table++;

			fs->metadirty = 1;
			ext2_sync(fs);

			if (ext2_is_group_sparse(fs, i))
			{
				ext2_copy_block(fs, fs->gd[i].bg_inode_bitmap, fs->gd[i].bg_inode_bitmap+1);
				fs->gd[i].bg_inode_bitmap++;

				fs->metadirty = 1;
				ext2_sync(fs);

				ext2_copy_block(fs, fs->gd[i].bg_block_bitmap, fs->gd[i].bg_block_bitmap+1);
				fs->gd[i].bg_block_bitmap++;

				fs->metadirty = 1;
				ext2_sync(fs);
			}
			else
				ext2_set_block_state(fs, fs->gd[i].bg_inode_table-1, 0, 1);
		}

		fs->gdblocks = newgdblocks;
		fs->metadirty = 1;
		ext2_sync(fs);

		fs->adminblocks++;
		if (sparse)
			admin++;
	}

	fs->numgroups++;

	fs->metadirty = 1;

	fs->sb.s_inodes_count += fs->sb.s_inodes_per_group;
	fs->sb.s_blocks_count += groupsize;
	fs->sb.s_free_blocks_count += groupsize - admin;
	fs->sb.s_free_inodes_count += fs->sb.s_inodes_per_group;

	{
		int off;
		int sparseoff;

		off = groupstart;
		sparseoff = off + fs->gdblocks + 1;

		if (sparse)
		{
			fs->gd[group].bg_block_bitmap = sparseoff;
			fs->gd[group].bg_inode_bitmap = sparseoff + 1;
		}
		else
		{
			fs->gd[group].bg_block_bitmap = off;
			fs->gd[group].bg_inode_bitmap = off + 1;
		}

		fs->gd[group].bg_inode_table = sparseoff + 2;			/* Hey, I don't know _why_ either */
	}

	fs->gd[group].bg_free_blocks_count = groupsize - admin;
	fs->gd[group].bg_free_inodes_count = fs->sb.s_inodes_per_group;
	fs->gd[group].bg_used_dirs_count   = 0;

	{
		struct ext2_buffer_head *bh;

		bh = ext2_bcreate(fs, fs->gd[group].bg_block_bitmap);

		if (sparse)
		{
			bh->data[0] |= _bitmap[0];
			for (i=1;i<=fs->gdblocks;i++)
				bh->data[i>>3] |= _bitmap[i&7];
		}

		i = fs->gd[group].bg_block_bitmap - groupstart;
		bh->data[i>>3] |= _bitmap[i&7];

		i = fs->gd[group].bg_inode_bitmap - groupstart;
		bh->data[i>>3] |= _bitmap[i&7];

		for (i=0;i<fs->inodeblocks;i++)
		{
			blk_t j;

			j = fs->gd[group].bg_inode_table - groupstart + i;
			bh->data[j>>3] |= _bitmap[j&7];
		}

		for (i=groupsize;i<fs->sb.s_blocks_per_group;i++)
			bh->data[i>>3] |= _bitmap[i&7];

		ext2_brelse(bh, 0);         /* this is a block bitmap */
	}

	ext2_zero_block(fs, fs->gd[group].bg_inode_bitmap);
	for (i=0;i<fs->inodeblocks;i++)
		ext2_zero_block(fs, fs->gd[group].bg_inode_table + i);
}

static void ext2_del_group(struct ext2_fs *fs)
{
	blk_t admin;
	int   group;
	blk_t groupsize;
	blk_t newgdblocks;
	int   sparse;

	fprintf(stderr, "ext2_del_group\n");

	group = fs->numgroups - 1;
	sparse = ext2_is_group_sparse(fs, group);

	admin = fs->adminblocks;
	if (!sparse)
		admin -= fs->gdblocks + 1;

	groupsize = fs->sb.s_blocks_count - fs->sb.s_first_data_block - group * fs->sb.s_blocks_per_group;

	if (fs->sb.s_free_blocks_count < groupsize - admin)
	{
		fprintf(stderr, "ext2_del_group: filesystem is too occupied to remove a group!\n");
		return;
	}

	if (fs->sb.s_free_inodes_count < fs->sb.s_inodes_per_group)
	{
		fprintf(stderr, "ext2_del_group: filesystem has too many allocated inodes to remove a group!\n");
		return;
	}

	if (fs->gd[group].bg_free_inodes_count != fs->sb.s_inodes_per_group)
	{
		fprintf(stderr, "ext2_del_group: this should not happen anymore!\n");
		return;
	}

	newgdblocks = howmany((fs->numgroups - 1) * sizeof(struct ext2_group_desc), fs->blocksize);
	if (newgdblocks != fs->gdblocks)
	{
		int i;

		fs->gdblocks = newgdblocks;

		for (i=0;i<fs->numgroups;i++)
		{
			blk_t block;
			blk_t num;
			blk_t j;
			int   sparse;

			sparse = ext2_is_group_sparse(fs, i);

			if (sparse)
			{
				block = fs->gd[i].bg_block_bitmap;
				num = fs->inodeblocks + 2;
			}
			else
			{
				block = fs->gd[i].bg_inode_table;
				num = fs->inodeblocks;

				if (ext2_get_block_state(fs, block-1))
				{
					fprintf(stderr, "error: block relocator should have relocated %i\n", block-1);
					return;
				}
			}

			for (j=0;j<num;j++)
				ext2_copy_block(fs, block+j, block+j-1);

			if (sparse)
				fs->gd[i].bg_block_bitmap--;

			ext2_set_block_state(fs, block+num-1, 0, 1);
			fs->gd[i].bg_inode_table--;

			if (sparse)
				fs->gd[i].bg_inode_bitmap--;
			else
				ext2_set_block_state(fs, block-1, 1, 1);

			fs->metadirty = 1;
			ext2_sync(fs);
		}

		fs->adminblocks--;
		if (sparse)
			admin--;
	}

	if (fs->gd[group].bg_free_blocks_count != groupsize - admin)
	{
		blk_t i;
		blk_t num;
		blk_t offset;

		offset = fs->sb.s_first_data_block + group * fs->sb.s_blocks_per_group;
		num = fs->sb.s_blocks_per_group;

		for (i=0;i<num;i++)
			if (ext2_is_data_block(fs, offset+i) && ext2_get_block_state(fs, offset+i))
			{
				fprintf(stderr, "error: block relocator should have relocated %i\n", offset+i);
				return;
			}
	}

	fs->numgroups--;

	fs->sb.s_inodes_count -= fs->sb.s_inodes_per_group;
	fs->sb.s_blocks_count -= groupsize;
	fs->sb.s_free_blocks_count -= groupsize - admin;
	fs->sb.s_free_inodes_count -= fs->sb.s_inodes_per_group;

	fs->metadirty = 1;
	ext2_sync(fs);
}

void ext2_grow_group(struct ext2_fs *fs, blk_t newsize)
{
	int   group;
	blk_t groupoff;
	blk_t gblocks;
	blk_t i;

	fprintf(stderr, "ext2_grow_group\n");

	group = fs->numgroups - 1;
	groupoff = group * fs->sb.s_blocks_per_group + fs->sb.s_first_data_block;
	gblocks = fs->sb.s_blocks_count - groupoff;

	if (newsize < gblocks)
	{
		fprintf(stderr, "ext2_grow_group: called to shrink group!\n");
		return;
	}

	if (gblocks == newsize)
	{
		fprintf(stderr, "ext2_grow_group: nothing to do!\n");
		return;
	}

	for (i=gblocks;i<newsize;i++)
		ext2_set_block_state(fs, groupoff + i, 0, 1);

	fs->sb.s_blocks_count += newsize - gblocks;

	fs->metadirty = 1;
	ext2_sync(fs);
}

void ext2_shrink_group(struct ext2_fs *fs, blk_t newsize)
{
	blk_t admin;
	int   group;
	blk_t groupoff;
	blk_t gblocks;
	int   i;

	fprintf(stderr, "ext2_shrink_group\n");

	group = fs->numgroups - 1;

	admin = fs->adminblocks;
	if (!ext2_is_group_sparse(fs, group))
		admin -= fs->gdblocks + 1;

	groupoff = group * fs->sb.s_blocks_per_group + fs->sb.s_first_data_block;
	gblocks = fs->sb.s_blocks_count - groupoff;

	if (newsize < admin)
	{
		fprintf(stderr, "ext2_shrink_group: cant shrink a group to %i blocks\n", newsize);
		return;
	}

	if (newsize > gblocks)
	{
		fprintf(stderr, "ext2_shrink_group: called to grow group!\n");
		return;
	}

	if (gblocks == newsize)
	{
		fprintf(stderr, "ext2_shrink_group: nothing to do!\n");
		return;
	}

	for (i=newsize;i<gblocks;i++)
	{
		if (ext2_get_block_state(fs, groupoff + i))
		{
			fprintf(stderr, "error: block relocator should have relocated %i\n", groupoff + i);
			return;
		}

		ext2_set_block_state(fs, groupoff + i, 1, 0);
	}

	i = gblocks - newsize;
	fs->metadirty = 1;
	fs->sb.s_blocks_count -= i;
	fs->sb.s_free_blocks_count -= i;
	fs->gd[group].bg_free_blocks_count -= i;

	fs->metadirty = 1;
	ext2_sync(fs);
}






static int ext2_grow_fs(struct ext2_fs *fs, blk_t newsize)
{
	blk_t diff;
	blk_t sizelast;

	fprintf(stderr, "ext2_grow_fs\n");

	if (!ext2_block_relocate(fs, newsize))
		return 0;

	diff = newsize - fs->sb.s_blocks_count;
	sizelast = fs->sb.s_blocks_count - fs->sb.s_first_data_block - (fs->numgroups-1)*fs->sb.s_blocks_per_group;

	if (sizelast != fs->sb.s_blocks_per_group)
	{
		blk_t growto;

		growto = sizelast + diff;
		if (growto > fs->sb.s_blocks_per_group)
			growto = fs->sb.s_blocks_per_group;

		ext2_grow_group(fs, growto);
		diff -= growto - sizelast;
	}

	while (diff)
	{
		sizelast = min(diff, fs->sb.s_blocks_per_group);
		ext2_add_group(fs, sizelast);
		diff -= sizelast;
	}

	return 1;
}

static int ext2_shrink_fs(struct ext2_fs *fs, blk_t newsize)
{
	blk_t diff;
	int newgroups;
	blk_t sizelast;

	fprintf(stderr, "ext2_shrink_fs\n");

	newgroups = howmany(newsize - fs->sb.s_first_data_block, fs->sb.s_blocks_per_group);
	if (fs->sb.s_blocks_count - fs->sb.s_free_blocks_count > newsize)
	{
		fprintf(stderr, "Your filesystem is too occupied to resize it to %i\n", newsize);
		fprintf(stderr, "blocks. Sorry.\n");
		return 0;
	}

	if (fs->sb.s_inodes_count - fs->sb.s_free_inodes_count > newgroups * fs->sb.s_inodes_per_group)
	{
		fprintf(stderr, "Your filesystem has too much occupied inodes to resize it\n");
		fprintf(stderr, "to %i blocks. Sorry.\n", newsize);
		return 0;
	}

	if (!ext2_inode_relocate(fs, newgroups))
		return 0;

	if (!ext2_block_relocate(fs, newsize))
		return 0;

	diff = fs->sb.s_blocks_count - newsize;

	while (diff > 0)
	{
		sizelast = fs->sb.s_blocks_count - fs->sb.s_first_data_block -
			(fs->numgroups - 1) * fs->sb.s_blocks_per_group;

		if (diff < sizelast)
		{
			ext2_shrink_group(fs, sizelast - diff);
			diff = 0;
		}
		else
		{
			ext2_del_group(fs);
			diff -= sizelast;
		}
	}

	return 1;
}





int ext2_resize_fs(struct ext2_fs *fs, blk_t newsize)
{
	blk_t container;
	blk_t residue;

	fprintf(stderr, "ext2_resize_fs\n");

	if (newsize == fs->sb.s_blocks_count)
		return 1;

	container = fs->handle->ops->get_size(fs->handle->cookie);
	if (newsize > container)
	{
		fprintf(stderr, "ext2_resize_fs: your container (i.e. the file or partition\n");
		fprintf(stderr, "                the filesystem) is in is not big enough for\n");
		fprintf(stderr, "                a filesystem of %i blocks (it's only\n", newsize);
		fprintf(stderr, "                %i blocks)\n", container);
		return 0;
	}

	residue = (newsize - fs->sb.s_first_data_block) % fs->sb.s_blocks_per_group;
	if (residue && residue <= fs->adminblocks)
	{
		fprintf(stderr, "ext2_resize_fs: %i is an impossible size for an ext2 fs! using %i instead\n",
			newsize, newsize-residue);
		newsize -= residue;
	}

	if (newsize < fs->sb.s_blocks_count)
		return ext2_shrink_fs(fs, newsize);

	return ext2_grow_fs(fs, newsize);
}
